@@ -2,8 +2,8 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2023 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2023 Laszlo Molnar
+   Copyright (C) 1996-2024 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2024 Laszlo Molnar
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -26,47 +26,120 @@
  */
 
 #pragma once
-#ifndef UPX_BELE_H__
-#define UPX_BELE_H__ 1
 
 // BE - Big Endian
 // LE - Little Endian
-// NE - Native Endianness (aka host endianness)
+// NE - Native Endianness (aka Host Endianness aka CPU Endianness)
 // TE - Target Endianness (not used here, see various packers)
 
+#if 1
+// some platforms may provide their own system bswapXX() functions, so rename to avoid conflicts
+#undef bswap16
+#undef bswap32
+#undef bswap64
+#define bswap16 upx_bswap16
+#define bswap32 upx_bswap32
+#define bswap64 upx_bswap64
+#endif
+
+/*************************************************************************
+// XE - eXtended Endian compatibility
+// try to detect XX16 vs XX32 vs XX64 size mismatches
+**************************************************************************/
+
+#if !(DEBUG)
+
+// permissive version using "void *"
+#define REQUIRE_XE16 /*empty*/
+#define REQUIRE_XE24 /*empty*/
+#define REQUIRE_XE32 /*empty*/
+#define REQUIRE_XE64 /*empty*/
+typedef void XE16;
+typedef void XE24;
+typedef void XE32;
+typedef void XE64;
+
+#else // permissive version
+
+// forward declarations
+struct BE16;
+struct BE32;
+struct BE64;
+struct LE16;
+struct LE32;
+struct LE64;
+
+// Note:
+//   void is explicitly allowed (but there is no automatic pointer conversion because of template!)
+//   char is explicitly allowed
+//   byte is explicitly allowed
+template <class T>
+inline constexpr bool is_xe16_type =
+    upx::is_same_any_v<T, void, char, byte, upx_int16_t, upx_uint16_t, BE16, LE16>;
+template <class T>
+inline constexpr bool is_xe24_type = upx::is_same_any_v<T, void, char, byte>;
+template <class T>
+inline constexpr bool is_xe32_type =
+    upx::is_same_any_v<T, void, char, byte, upx_int32_t, upx_uint32_t, BE32, LE32>;
+template <class T>
+inline constexpr bool is_xe64_type =
+    upx::is_same_any_v<T, void, char, byte, upx_int64_t, upx_uint64_t, BE64, LE64>;
+
+template <class T>
+using enable_if_xe16 = std::enable_if_t<is_xe16_type<T>, T>;
+template <class T>
+using enable_if_xe24 = std::enable_if_t<is_xe24_type<T>, T>;
+template <class T>
+using enable_if_xe32 = std::enable_if_t<is_xe32_type<T>, T>;
+template <class T>
+using enable_if_xe64 = std::enable_if_t<is_xe64_type<T>, T>;
+
+#define REQUIRE_XE16 template <class XE16, class = enable_if_xe16<XE16> >
+#define REQUIRE_XE24 template <class XE24, class = enable_if_xe24<XE24> >
+#define REQUIRE_XE32 template <class XE32, class = enable_if_xe32<XE32> >
+#define REQUIRE_XE64 template <class XE64, class = enable_if_xe64<XE64> >
+
+#endif // permissive version
+
 /*************************************************************************
 // core - NE
 **************************************************************************/
 
-static forceinline unsigned get_ne16(const void *p) {
+REQUIRE_XE16
+static forceinline unsigned get_ne16(const XE16 *p) noexcept {
     upx_uint16_t v = 0;
     upx_memcpy_inline(&v, p, sizeof(v));
     return v;
 }
 
-static forceinline unsigned get_ne32(const void *p) {
+REQUIRE_XE32
+static forceinline unsigned get_ne32(const XE32 *p) noexcept {
     upx_uint32_t v = 0;
     upx_memcpy_inline(&v, p, sizeof(v));
     return v;
 }
 
-static forceinline upx_uint64_t get_ne64(const void *p) {
+REQUIRE_XE64
+static forceinline upx_uint64_t get_ne64(const XE64 *p) noexcept {
     upx_uint64_t v = 0;
     upx_memcpy_inline(&v, p, sizeof(v));
     return v;
 }
 
-static forceinline void set_ne16(void *p, unsigned vv) {
+REQUIRE_XE16
+static forceinline void set_ne16(XE16 *p, unsigned vv) noexcept {
     upx_uint16_t v = (upx_uint16_t) (vv & 0xffff);
     upx_memcpy_inline(p, &v, sizeof(v));
 }
 
-static forceinline void set_ne32(void *p, unsigned vv) {
+REQUIRE_XE32
+static forceinline void set_ne32(XE32 *p, unsigned vv) noexcept {
     upx_uint32_t v = vv;
     upx_memcpy_inline(p, &v, sizeof(v));
 }
 
-static forceinline void set_ne64(void *p, upx_uint64_t vv) {
+REQUIRE_XE64
+static forceinline void set_ne64(XE64 *p, upx_uint64_t vv) noexcept {
     upx_uint64_t v = vv;
     upx_memcpy_inline(p, &v, sizeof(v));
 }
@@ -79,31 +152,46 @@ static forceinline void set_ne64(void *p, upx_uint64_t vv) {
 
 ACC_COMPILE_TIME_ASSERT_HEADER(sizeof(long) == 4)
 
-// unfortunately *not* constexpr with MSVC
-static forceinline unsigned bswap16(unsigned v) { return (unsigned) _byteswap_ulong(v << 16); }
-static forceinline unsigned bswap32(unsigned v) { return (unsigned) _byteswap_ulong(v); }
-static forceinline upx_uint64_t bswap64(upx_uint64_t v) { return _byteswap_uint64(v); }
+// unfortunately *not* constexpr with current MSVC
+static forceinline /*constexpr*/ unsigned bswap16(unsigned v) noexcept {
+    return (unsigned) _byteswap_ulong(v << 16);
+}
+static forceinline /*constexpr*/ unsigned bswap32(unsigned v) noexcept {
+    return (unsigned) _byteswap_ulong(v);
+}
+static forceinline /*constexpr*/ upx_uint64_t bswap64(upx_uint64_t v) noexcept {
+    return _byteswap_uint64(v);
+}
 
 #else
 
-static forceinline constexpr unsigned bswap16(unsigned v) {
+static forceinline constexpr unsigned bswap16(unsigned v) noexcept {
+#if defined(__riscv) && __riscv_xlen == 64
+    return (unsigned) __builtin_bswap64((upx_uint64_t) v << 48);
+#else
     // return __builtin_bswap16((upx_uint16_t) (v & 0xffff));
-    // return (unsigned) __builtin_bswap64((upx_uint64_t) v << 48);
     return __builtin_bswap32(v << 16);
+#endif
 }
-static forceinline constexpr unsigned bswap32(unsigned v) {
-    // return (unsigned) __builtin_bswap64((upx_uint64_t) v << 32);
+static forceinline constexpr unsigned bswap32(unsigned v) noexcept {
+#if defined(__riscv) && __riscv_xlen == 64
+    return (unsigned) __builtin_bswap64((upx_uint64_t) v << 32);
+#else
     return __builtin_bswap32(v);
+#endif
+}
+static forceinline constexpr upx_uint64_t bswap64(upx_uint64_t v) noexcept {
+    return __builtin_bswap64(v);
 }
-static forceinline constexpr upx_uint64_t bswap64(upx_uint64_t v) { return __builtin_bswap64(v); }
 
 #endif
 
-static forceinline constexpr unsigned no_bswap16(unsigned v) {
-    return v & 0xffff; // needed so that this is equivalent to bswap16() above
+static forceinline constexpr unsigned no_bswap16(unsigned v) noexcept {
+    // mask is needed so that for all v: bswap16(bswap16(v)) == no_bswap16(v)
+    return v & 0xffff;
 }
-static forceinline constexpr unsigned no_bswap32(unsigned v) { return v; }
-static forceinline constexpr upx_uint64_t no_bswap64(upx_uint64_t v) { return v; }
+static forceinline constexpr unsigned no_bswap32(unsigned v) noexcept { return v; }
+static forceinline constexpr upx_uint64_t no_bswap64(upx_uint64_t v) noexcept { return v; }
 
 #if (ACC_ABI_BIG_ENDIAN)
 #define ne16_to_be16(v) no_bswap16(v)
@@ -125,114 +213,154 @@ static forceinline constexpr upx_uint64_t no_bswap64(upx_uint64_t v) { return v;
 // get/set 16/32/64
 **************************************************************************/
 
-inline unsigned get_be16(const void *p) { return ne16_to_be16(get_ne16(p)); }
-inline unsigned get_be32(const void *p) { return ne32_to_be32(get_ne32(p)); }
-inline upx_uint64_t get_be64(const void *p) { return ne64_to_be64(get_ne64(p)); }
-inline unsigned get_le16(const void *p) { return ne16_to_le16(get_ne16(p)); }
-inline unsigned get_le32(const void *p) { return ne32_to_le32(get_ne32(p)); }
-inline upx_uint64_t get_le64(const void *p) { return ne64_to_le64(get_ne64(p)); }
-inline void set_be16(void *p, unsigned v) { set_ne16(p, ne16_to_be16(v)); }
-inline void set_be32(void *p, unsigned v) { set_ne32(p, ne32_to_be32(v)); }
-inline void set_be64(void *p, upx_uint64_t v) { set_ne64(p, ne64_to_be64(v)); }
-inline void set_le16(void *p, unsigned v) { set_ne16(p, ne16_to_le16(v)); }
-inline void set_le32(void *p, unsigned v) { set_ne32(p, ne32_to_le32(v)); }
-inline void set_le64(void *p, upx_uint64_t v) { set_ne64(p, ne64_to_le64(v)); }
+REQUIRE_XE16
+inline unsigned get_be16(const XE16 *p) noexcept { return ne16_to_be16(get_ne16(p)); }
+REQUIRE_XE32
+inline unsigned get_be32(const XE32 *p) noexcept { return ne32_to_be32(get_ne32(p)); }
+REQUIRE_XE64
+inline upx_uint64_t get_be64(const XE64 *p) noexcept { return ne64_to_be64(get_ne64(p)); }
+REQUIRE_XE16
+inline unsigned get_le16(const XE16 *p) noexcept { return ne16_to_le16(get_ne16(p)); }
+REQUIRE_XE32
+inline unsigned get_le32(const XE32 *p) noexcept { return ne32_to_le32(get_ne32(p)); }
+REQUIRE_XE64
+inline upx_uint64_t get_le64(const XE64 *p) noexcept { return ne64_to_le64(get_ne64(p)); }
+
+REQUIRE_XE16
+inline void set_be16(XE16 *p, unsigned v) noexcept { set_ne16(p, ne16_to_be16(v)); }
+REQUIRE_XE32
+inline void set_be32(XE32 *p, unsigned v) noexcept { set_ne32(p, ne32_to_be32(v)); }
+REQUIRE_XE64
+inline void set_be64(XE64 *p, upx_uint64_t v) noexcept { set_ne64(p, ne64_to_be64(v)); }
+REQUIRE_XE16
+inline void set_le16(XE16 *p, unsigned v) noexcept { set_ne16(p, ne16_to_le16(v)); }
+REQUIRE_XE32
+inline void set_le32(XE32 *p, unsigned v) noexcept { set_ne32(p, ne32_to_le32(v)); }
+REQUIRE_XE64
+inline void set_le64(XE64 *p, upx_uint64_t v) noexcept { set_ne64(p, ne64_to_le64(v)); }
 
 /*************************************************************************
 // get/set 24/26
 **************************************************************************/
 
-inline unsigned get_be24(const void *p) {
-    const unsigned char *b = ACC_CCAST(const unsigned char *, p);
+REQUIRE_XE24
+inline unsigned get_be24(const XE24 *p) noexcept {
+    const byte *b = ACC_CCAST(const byte *, p);
     return (b[0] << 16) | (b[1] << 8) | (b[2] << 0);
 }
 
-inline unsigned get_le24(const void *p) {
-    const unsigned char *b = ACC_CCAST(const unsigned char *, p);
+REQUIRE_XE24
+inline unsigned get_le24(const XE24 *p) noexcept {
+    const byte *b = ACC_CCAST(const byte *, p);
     return (b[0] << 0) | (b[1] << 8) | (b[2] << 16);
 }
 
-inline void set_be24(void *p, unsigned v) {
-    unsigned char *b = ACC_PCAST(unsigned char *, p);
-    b[0] = ACC_ICONV(unsigned char, (v >> 16) & 0xff);
-    b[1] = ACC_ICONV(unsigned char, (v >> 8) & 0xff);
-    b[2] = ACC_ICONV(unsigned char, (v >> 0) & 0xff);
+REQUIRE_XE24
+inline void set_be24(XE24 *p, unsigned v) noexcept {
+    byte *b = ACC_PCAST(byte *, p);
+    b[0] = ACC_ICONV(byte, (v >> 16) & 0xff);
+    b[1] = ACC_ICONV(byte, (v >> 8) & 0xff);
+    b[2] = ACC_ICONV(byte, (v >> 0) & 0xff);
 }
 
-inline void set_le24(void *p, unsigned v) {
-    unsigned char *b = ACC_PCAST(unsigned char *, p);
-    b[0] = ACC_ICONV(unsigned char, (v >> 0) & 0xff);
-    b[1] = ACC_ICONV(unsigned char, (v >> 8) & 0xff);
-    b[2] = ACC_ICONV(unsigned char, (v >> 16) & 0xff);
+REQUIRE_XE24
+inline void set_le24(XE24 *p, unsigned v) noexcept {
+    byte *b = ACC_PCAST(byte *, p);
+    b[0] = ACC_ICONV(byte, (v >> 0) & 0xff);
+    b[1] = ACC_ICONV(byte, (v >> 8) & 0xff);
+    b[2] = ACC_ICONV(byte, (v >> 16) & 0xff);
 }
 
-inline unsigned get_le26(const void *p) { return get_le32(p) & 0x03ffffff; }
+REQUIRE_XE32
+inline unsigned get_le26(const XE32 *p) noexcept { return get_le32(p) & 0x03ffffff; }
+REQUIRE_XE32
+inline unsigned get_le19_5(const XE32 *p) noexcept { return (get_le32(p) >> 5) & 0x0007ffff; }
+REQUIRE_XE32
+inline unsigned get_le14_5(const XE32 *p) noexcept { return (get_le32(p) >> 5) & 0x00003fff; }
 
-inline void set_le26(void *p, unsigned v) {
+REQUIRE_XE32
+inline void set_le26(XE32 *p, unsigned v) noexcept {
     // preserve the top 6 bits
-#if 0
-    set_le32(p, (get_le32(p) & 0xfc000000) | (v & 0x03ffffff));
-#else
+    //   set_le32(p, (get_le32(p) & 0xfc000000) | (v & 0x03ffffff));
     // optimized version, saving a runtime bswap32
     set_ne32(p, (get_ne32(p) & ne32_to_le32(0xfc000000)) |
                     (ne32_to_le32(v) & ne32_to_le32(0x03ffffff)));
-#endif
+}
+REQUIRE_XE32
+inline void set_le19_5(XE32 *p, unsigned v) noexcept {
+    set_le32(p, (get_le32(p) & 0xff00001f) | ((v & 0x0007ffff) << 5));
+}
+REQUIRE_XE32
+inline void set_le14_5(XE32 *p, unsigned v) noexcept {
+    set_le32(p, (get_le32(p) & 0xfff8001f) | ((v & 0x00003fff) << 5));
 }
 
 /*************************************************************************
 // get signed values
 **************************************************************************/
 
-static forceinline int sign_extend(unsigned v, unsigned bits) {
+static forceinline constexpr int sign_extend(unsigned v, unsigned bits) noexcept {
+#if (ACC_ARCH_M68K) // no barrel shifter
     const unsigned sign_bit = 1u << (bits - 1);
-    v &= sign_bit | (sign_bit - 1);
-    v |= 0 - (v & sign_bit);
-    return ACC_ICAST(int, v);
+    return ACC_ICAST(int, (v & (sign_bit - 1)) - (v & sign_bit));
+#else
+    return ACC_ICAST(int, v << (32 - bits)) >> (32 - bits);
+#endif
 }
 
-static forceinline upx_int64_t sign_extend(upx_uint64_t v, unsigned bits) {
+static forceinline constexpr upx_int64_t sign_extend(upx_uint64_t v, unsigned bits) noexcept {
+#if (ACC_ARCH_M68K) // no barrel shifter
     const upx_uint64_t sign_bit = 1ull << (bits - 1);
-    v &= sign_bit | (sign_bit - 1);
-    v |= 0 - (v & sign_bit);
-    return ACC_ICAST(upx_int64_t, v);
+    return ACC_ICAST(upx_int64_t, (v & (sign_bit - 1)) - (v & sign_bit));
+#else
+    return ACC_ICAST(upx_int64_t, v << (64 - bits)) >> (64 - bits);
+#endif
 }
 
-inline int get_be16_signed(const void *p) {
+REQUIRE_XE16
+inline int get_be16_signed(const XE16 *p) noexcept {
     unsigned v = get_be16(p);
     return sign_extend(v, 16);
 }
 
-inline int get_be24_signed(const void *p) {
+REQUIRE_XE24
+inline int get_be24_signed(const XE24 *p) noexcept {
     unsigned v = get_be24(p);
     return sign_extend(v, 24);
 }
 
-inline int get_be32_signed(const void *p) {
+REQUIRE_XE32
+inline int get_be32_signed(const XE32 *p) noexcept {
     unsigned v = get_be32(p);
     return sign_extend(v, 32);
 }
 
-inline upx_int64_t get_be64_signed(const void *p) {
+REQUIRE_XE64
+inline upx_int64_t get_be64_signed(const XE64 *p) noexcept {
     upx_uint64_t v = get_be64(p);
     return sign_extend(v, 64);
 }
 
-inline int get_le16_signed(const void *p) {
+REQUIRE_XE16
+inline int get_le16_signed(const XE16 *p) noexcept {
     unsigned v = get_le16(p);
     return sign_extend(v, 16);
 }
 
-inline int get_le24_signed(const void *p) {
+REQUIRE_XE24
+inline int get_le24_signed(const XE24 *p) noexcept {
     unsigned v = get_le24(p);
     return sign_extend(v, 24);
 }
 
-inline int get_le32_signed(const void *p) {
+REQUIRE_XE32
+inline int get_le32_signed(const XE32 *p) noexcept {
     unsigned v = get_le32(p);
     return sign_extend(v, 32);
 }
 
-inline upx_int64_t get_le64_signed(const void *p) {
+REQUIRE_XE64
+inline upx_int64_t get_le64_signed(const XE64 *p) noexcept {
     upx_uint64_t v = get_le64(p);
     return sign_extend(v, 64);
 }
@@ -248,368 +376,380 @@ inline upx_int64_t get_le64_signed(const void *p) {
 //  to have gcc bug 17519 fixed - see http://gcc.gnu.org/PR17519 ]
 **************************************************************************/
 
-struct alignas(1) BE16 {
+struct alignas(1) BE16 final {
     typedef unsigned integral_conversion_type; // automatic conversion to unsigned
-    unsigned char d[2];
+    byte d[2];
 
-    BE16 &operator=(unsigned v) {
+    forceinline operator unsigned() const noexcept { return get_be16(d); }
+
+    forceinline BE16 &operator=(unsigned v) noexcept {
         set_be16(d, v);
         return *this;
     }
-    BE16 &operator+=(unsigned v) {
+    BE16 &operator+=(unsigned v) noexcept {
         set_be16(d, get_be16(d) + v);
         return *this;
     }
-    BE16 &operator-=(unsigned v) {
+    BE16 &operator-=(unsigned v) noexcept {
         set_be16(d, get_be16(d) - v);
         return *this;
     }
-    BE16 &operator*=(unsigned v) {
+    BE16 &operator*=(unsigned v) noexcept {
         set_be16(d, get_be16(d) * v);
         return *this;
     }
-    BE16 &operator/=(unsigned v) {
+    BE16 &operator/=(unsigned v) noexcept {
         set_be16(d, get_be16(d) / v);
         return *this;
     }
-    BE16 &operator&=(unsigned v) {
+    BE16 &operator&=(unsigned v) noexcept {
         set_be16(d, get_be16(d) & v);
         return *this;
     }
-    BE16 &operator|=(unsigned v) {
+    BE16 &operator|=(unsigned v) noexcept {
         set_be16(d, get_be16(d) | v);
         return *this;
     }
-    BE16 &operator^=(unsigned v) {
+    BE16 &operator^=(unsigned v) noexcept {
         set_be16(d, get_be16(d) ^ v);
         return *this;
     }
-    BE16 &operator<<=(unsigned v) {
+    BE16 &operator<<=(unsigned v) noexcept {
         set_be16(d, get_be16(d) << v);
         return *this;
     }
-    BE16 &operator>>=(unsigned v) {
+    BE16 &operator>>=(unsigned v) noexcept {
         set_be16(d, get_be16(d) >> v);
         return *this;
     }
 
-    operator unsigned() const { return get_be16(d); }
-
-    bool operator<(const BE16 &v) const { return unsigned(*this) < unsigned(v); }
+    bool operator==(const BE16 &x) const noexcept { return memcmp(d, x.d, sizeof(d)) == 0; }
+    bool operator<(const BE16 &x) const noexcept { return unsigned(*this) < unsigned(x); }
 };
 
-struct alignas(1) BE32 {
+struct alignas(1) BE32 final {
     typedef unsigned integral_conversion_type; // automatic conversion to unsigned
-    unsigned char d[4];
+    byte d[4];
 
-    BE32 &operator=(unsigned v) {
+    forceinline operator unsigned() const noexcept { return get_be32(d); }
+
+    forceinline BE32 &operator=(unsigned v) noexcept {
         set_be32(d, v);
         return *this;
     }
-    BE32 &operator+=(unsigned v) {
+    BE32 &operator+=(unsigned v) noexcept {
         set_be32(d, get_be32(d) + v);
         return *this;
     }
-    BE32 &operator-=(unsigned v) {
+    BE32 &operator-=(unsigned v) noexcept {
         set_be32(d, get_be32(d) - v);
         return *this;
     }
-    BE32 &operator*=(unsigned v) {
+    BE32 &operator*=(unsigned v) noexcept {
         set_be32(d, get_be32(d) * v);
         return *this;
     }
-    BE32 &operator/=(unsigned v) {
+    BE32 &operator/=(unsigned v) noexcept {
         set_be32(d, get_be32(d) / v);
         return *this;
     }
-    BE32 &operator&=(unsigned v) {
+    BE32 &operator&=(unsigned v) noexcept {
         set_be32(d, get_be32(d) & v);
         return *this;
     }
-    BE32 &operator|=(unsigned v) {
+    BE32 &operator|=(unsigned v) noexcept {
         set_be32(d, get_be32(d) | v);
         return *this;
     }
-    BE32 &operator^=(unsigned v) {
+    BE32 &operator^=(unsigned v) noexcept {
         set_be32(d, get_be32(d) ^ v);
         return *this;
     }
-    BE32 &operator<<=(unsigned v) {
+    BE32 &operator<<=(unsigned v) noexcept {
         set_be32(d, get_be32(d) << v);
         return *this;
     }
-    BE32 &operator>>=(unsigned v) {
+    BE32 &operator>>=(unsigned v) noexcept {
         set_be32(d, get_be32(d) >> v);
         return *this;
     }
 
-    operator unsigned() const { return get_be32(d); }
-
-    bool operator<(const BE32 &v) const { return unsigned(*this) < unsigned(v); }
+    bool operator==(const BE32 &x) const noexcept { return memcmp(d, x.d, sizeof(d)) == 0; }
+    bool operator<(const BE32 &x) const noexcept { return unsigned(*this) < unsigned(x); }
 };
 
-struct alignas(1) BE64 {
+struct alignas(1) BE64 final {
     typedef upx_uint64_t integral_conversion_type; // automatic conversion to upx_uint64_t
-    unsigned char d[8];
+    byte d[8];
 
-    BE64 &operator=(upx_uint64_t v) {
+    forceinline operator upx_uint64_t() const noexcept { return get_be64(d); }
+
+    forceinline BE64 &operator=(upx_uint64_t v) noexcept {
         set_be64(d, v);
         return *this;
     }
-    BE64 &operator+=(upx_uint64_t v) {
+    BE64 &operator+=(upx_uint64_t v) noexcept {
         set_be64(d, get_be64(d) + v);
         return *this;
     }
-    BE64 &operator-=(upx_uint64_t v) {
+    BE64 &operator-=(upx_uint64_t v) noexcept {
         set_be64(d, get_be64(d) - v);
         return *this;
     }
-    BE64 &operator*=(upx_uint64_t v) {
+    BE64 &operator*=(upx_uint64_t v) noexcept {
         set_be64(d, get_be64(d) * v);
         return *this;
     }
-    BE64 &operator/=(upx_uint64_t v) {
+    BE64 &operator/=(upx_uint64_t v) noexcept {
         set_be64(d, get_be64(d) / v);
         return *this;
     }
-    BE64 &operator&=(upx_uint64_t v) {
+    BE64 &operator&=(upx_uint64_t v) noexcept {
         set_be64(d, get_be64(d) & v);
         return *this;
     }
-    BE64 &operator|=(upx_uint64_t v) {
+    BE64 &operator|=(upx_uint64_t v) noexcept {
         set_be64(d, get_be64(d) | v);
         return *this;
     }
-    BE64 &operator^=(upx_uint64_t v) {
+    BE64 &operator^=(upx_uint64_t v) noexcept {
         set_be64(d, get_be64(d) ^ v);
         return *this;
     }
-    BE64 &operator<<=(unsigned v) {
+    BE64 &operator<<=(unsigned v) noexcept {
         set_be64(d, get_be64(d) << v);
         return *this;
     }
-    BE64 &operator>>=(unsigned v) {
+    BE64 &operator>>=(unsigned v) noexcept {
         set_be64(d, get_be64(d) >> v);
         return *this;
     }
 
-    operator upx_uint64_t() const { return get_be64(d); }
-
-    bool operator<(const BE64 &v) const { return upx_uint64_t(*this) < upx_uint64_t(v); }
+    bool operator==(const BE64 &x) const noexcept { return memcmp(d, x.d, sizeof(d)) == 0; }
+    bool operator<(const BE64 &x) const noexcept { return upx_uint64_t(*this) < upx_uint64_t(x); }
 };
 
-struct alignas(1) LE16 {
+struct alignas(1) LE16 final {
     typedef unsigned integral_conversion_type; // automatic conversion to unsigned
-    unsigned char d[2];
+    byte d[2];
 
-    LE16 &operator=(unsigned v) {
+    forceinline operator unsigned() const noexcept { return get_le16(d); }
+
+    forceinline LE16 &operator=(unsigned v) noexcept {
         set_le16(d, v);
         return *this;
     }
-    LE16 &operator+=(unsigned v) {
+    LE16 &operator+=(unsigned v) noexcept {
         set_le16(d, get_le16(d) + v);
         return *this;
     }
-    LE16 &operator-=(unsigned v) {
+    LE16 &operator-=(unsigned v) noexcept {
         set_le16(d, get_le16(d) - v);
         return *this;
     }
-    LE16 &operator*=(unsigned v) {
+    LE16 &operator*=(unsigned v) noexcept {
         set_le16(d, get_le16(d) * v);
         return *this;
     }
-    LE16 &operator/=(unsigned v) {
+    LE16 &operator/=(unsigned v) noexcept {
         set_le16(d, get_le16(d) / v);
         return *this;
     }
-    LE16 &operator&=(unsigned v) {
+    LE16 &operator&=(unsigned v) noexcept {
         set_le16(d, get_le16(d) & v);
         return *this;
     }
-    LE16 &operator|=(unsigned v) {
+    LE16 &operator|=(unsigned v) noexcept {
         set_le16(d, get_le16(d) | v);
         return *this;
     }
-    LE16 &operator^=(unsigned v) {
+    LE16 &operator^=(unsigned v) noexcept {
         set_le16(d, get_le16(d) ^ v);
         return *this;
     }
-    LE16 &operator<<=(unsigned v) {
+    LE16 &operator<<=(unsigned v) noexcept {
         set_le16(d, get_le16(d) << v);
         return *this;
     }
-    LE16 &operator>>=(unsigned v) {
+    LE16 &operator>>=(unsigned v) noexcept {
         set_le16(d, get_le16(d) >> v);
         return *this;
     }
 
-    operator unsigned() const { return get_le16(d); }
-
-    bool operator<(const LE16 &v) const { return unsigned(*this) < unsigned(v); }
+    bool operator==(const LE16 &x) const noexcept { return memcmp(d, x.d, sizeof(d)) == 0; }
+    bool operator<(const LE16 &x) const noexcept { return unsigned(*this) < unsigned(x); }
 };
 
-struct alignas(1) LE32 {
+struct alignas(1) LE32 final {
     typedef unsigned integral_conversion_type; // automatic conversion to unsigned
-    unsigned char d[4];
+    byte d[4];
 
-    LE32 &operator=(unsigned v) {
+    forceinline operator unsigned() const noexcept { return get_le32(d); }
+
+    forceinline LE32 &operator=(unsigned v) noexcept {
         set_le32(d, v);
         return *this;
     }
-    LE32 &operator+=(unsigned v) {
+    LE32 &operator+=(unsigned v) noexcept {
         set_le32(d, get_le32(d) + v);
         return *this;
     }
-    LE32 &operator-=(unsigned v) {
+    LE32 &operator-=(unsigned v) noexcept {
         set_le32(d, get_le32(d) - v);
         return *this;
     }
-    LE32 &operator*=(unsigned v) {
+    LE32 &operator*=(unsigned v) noexcept {
         set_le32(d, get_le32(d) * v);
         return *this;
     }
-    LE32 &operator/=(unsigned v) {
+    LE32 &operator/=(unsigned v) noexcept {
         set_le32(d, get_le32(d) / v);
         return *this;
     }
-    LE32 &operator&=(unsigned v) {
+    LE32 &operator&=(unsigned v) noexcept {
         set_le32(d, get_le32(d) & v);
         return *this;
     }
-    LE32 &operator|=(unsigned v) {
+    LE32 &operator|=(unsigned v) noexcept {
         set_le32(d, get_le32(d) | v);
         return *this;
     }
-    LE32 &operator^=(unsigned v) {
+    LE32 &operator^=(unsigned v) noexcept {
         set_le32(d, get_le32(d) ^ v);
         return *this;
     }
-    LE32 &operator<<=(unsigned v) {
+    LE32 &operator<<=(unsigned v) noexcept {
         set_le32(d, get_le32(d) << v);
         return *this;
     }
-    LE32 &operator>>=(unsigned v) {
+    LE32 &operator>>=(unsigned v) noexcept {
         set_le32(d, get_le32(d) >> v);
         return *this;
     }
 
-    operator unsigned() const { return get_le32(d); }
-
-    bool operator<(const LE32 &v) const { return unsigned(*this) < unsigned(v); }
+    bool operator==(const LE32 &x) const noexcept { return memcmp(d, x.d, sizeof(d)) == 0; }
+    bool operator<(const LE32 &x) const noexcept { return unsigned(*this) < unsigned(x); }
 };
 
-struct alignas(1) LE64 {
+struct alignas(1) LE64 final {
     typedef upx_uint64_t integral_conversion_type; // automatic conversion to upx_uint64_t
-    unsigned char d[8];
+    byte d[8];
 
-    LE64 &operator=(upx_uint64_t v) {
+    forceinline operator upx_uint64_t() const noexcept { return get_le64(d); }
+
+    forceinline LE64 &operator=(upx_uint64_t v) noexcept {
         set_le64(d, v);
         return *this;
     }
-    LE64 &operator+=(upx_uint64_t v) {
+    LE64 &operator+=(upx_uint64_t v) noexcept {
         set_le64(d, get_le64(d) + v);
         return *this;
     }
-    LE64 &operator-=(upx_uint64_t v) {
+    LE64 &operator-=(upx_uint64_t v) noexcept {
         set_le64(d, get_le64(d) - v);
         return *this;
     }
-    LE64 &operator*=(upx_uint64_t v) {
+    LE64 &operator*=(upx_uint64_t v) noexcept {
         set_le64(d, get_le64(d) * v);
         return *this;
     }
-    LE64 &operator/=(upx_uint64_t v) {
+    LE64 &operator/=(upx_uint64_t v) noexcept {
         set_le64(d, get_le64(d) / v);
         return *this;
     }
-    LE64 &operator&=(upx_uint64_t v) {
+    LE64 &operator&=(upx_uint64_t v) noexcept {
         set_le64(d, get_le64(d) & v);
         return *this;
     }
-    LE64 &operator|=(upx_uint64_t v) {
+    LE64 &operator|=(upx_uint64_t v) noexcept {
         set_le64(d, get_le64(d) | v);
         return *this;
     }
-    LE64 &operator^=(upx_uint64_t v) {
+    LE64 &operator^=(upx_uint64_t v) noexcept {
         set_le64(d, get_le64(d) ^ v);
         return *this;
     }
-    LE64 &operator<<=(unsigned v) {
+    LE64 &operator<<=(unsigned v) noexcept {
         set_le64(d, get_le64(d) << v);
         return *this;
     }
-    LE64 &operator>>=(unsigned v) {
+    LE64 &operator>>=(unsigned v) noexcept {
         set_le64(d, get_le64(d) >> v);
         return *this;
     }
 
-    operator upx_uint64_t() const { return get_le64(d); }
-
-    bool operator<(const LE64 &v) const { return upx_uint64_t(*this) < upx_uint64_t(v); }
+    bool operator==(const LE64 &x) const noexcept { return memcmp(d, x.d, sizeof(d)) == 0; }
+    bool operator<(const LE64 &x) const noexcept { return upx_uint64_t(*this) < upx_uint64_t(x); }
 };
 
-// native types
-#if (ACC_ABI_BIG_ENDIAN)
-typedef BE16 NE16;
-typedef BE32 NE32;
-typedef BE64 NE64;
-#else
-typedef LE16 NE16;
-typedef LE32 NE32;
-typedef LE64 NE64;
-#endif
-
 /*************************************************************************
 // global operators (pointer addition/subtraction)
 **************************************************************************/
 
 template <class T>
-inline T *operator+(T *ptr, const BE16 &v) {
+inline T *operator+(T *ptr, const BE16 &v) noexcept {
     return ptr + unsigned(v);
 }
 template <class T>
-inline T *operator-(T *ptr, const BE16 &v) {
+inline T *operator-(T *ptr, const BE16 &v) noexcept {
     return ptr - unsigned(v);
 }
 template <class T>
-inline T *operator+(T *ptr, const BE32 &v) {
+inline T *operator+(T *ptr, const BE32 &v) noexcept {
     return ptr + unsigned(v);
 }
 template <class T>
-inline T *operator-(T *ptr, const BE32 &v) {
+inline T *operator-(T *ptr, const BE32 &v) noexcept {
     return ptr - unsigned(v);
 }
 template <class T>
-inline T *operator+(T *ptr, const LE16 &v) {
+inline T *operator+(T *ptr, const LE16 &v) noexcept {
     return ptr + unsigned(v);
 }
 template <class T>
-inline T *operator-(T *ptr, const LE16 &v) {
+inline T *operator-(T *ptr, const LE16 &v) noexcept {
     return ptr - unsigned(v);
 }
 template <class T>
-inline T *operator+(T *ptr, const LE32 &v) {
+inline T *operator+(T *ptr, const LE32 &v) noexcept {
     return ptr + unsigned(v);
 }
 template <class T>
-inline T *operator-(T *ptr, const LE32 &v) {
+inline T *operator-(T *ptr, const LE32 &v) noexcept {
     return ptr - unsigned(v);
 }
 
 // these are not implemented on purpose and will cause errors
 template <class T>
-T *operator+(T *ptr, const BE64 &v) DELETED_FUNCTION;
+T *operator+(T *ptr, const BE64 &v) noexcept DELETED_FUNCTION;
 template <class T>
-T *operator-(T *ptr, const BE64 &v) DELETED_FUNCTION;
+T *operator-(T *ptr, const BE64 &v) noexcept DELETED_FUNCTION;
 template <class T>
-T *operator+(T *ptr, const LE64 &v) DELETED_FUNCTION;
+T *operator+(T *ptr, const LE64 &v) noexcept DELETED_FUNCTION;
 template <class T>
-T *operator-(T *ptr, const LE64 &v) DELETED_FUNCTION;
+T *operator-(T *ptr, const LE64 &v) noexcept DELETED_FUNCTION;
+
+#if !ALLOW_INT_PLUS_MEMBUFFER
+template <class T>
+T *operator+(const BE16 &v, T *ptr) noexcept DELETED_FUNCTION;
+template <class T>
+T *operator+(const BE32 &v, T *ptr) noexcept DELETED_FUNCTION;
+template <class T>
+T *operator+(const LE16 &v, T *ptr) noexcept DELETED_FUNCTION;
+template <class T>
+T *operator+(const LE32 &v, T *ptr) noexcept DELETED_FUNCTION;
+#endif // ALLOW_INT_PLUS_MEMBUFFER
+
+template <class T>
+T *operator+(const BE64 &v, T *ptr) noexcept DELETED_FUNCTION;
+template <class T>
+T *operator+(const LE64 &v, T *ptr) noexcept DELETED_FUNCTION;
 
 /*************************************************************************
 // global overloads
 **************************************************************************/
 
+// TODO later: move these to upx namespace in util/cxxlib.h; see conf.h
 inline unsigned ALIGN_DOWN(unsigned a, const BE32 &b) { return ALIGN_DOWN(a, unsigned(b)); }
 inline unsigned ALIGN_DOWN(const BE32 &a, unsigned b) { return ALIGN_DOWN(unsigned(a), b); }
 inline unsigned ALIGN_UP(unsigned a, const BE32 &b) { return ALIGN_UP(a, unsigned(b)); }
@@ -620,6 +760,7 @@ inline unsigned ALIGN_DOWN(const LE32 &a, unsigned b) { return ALIGN_DOWN(unsign
 inline unsigned ALIGN_UP(unsigned a, const LE32 &b) { return ALIGN_UP(a, unsigned(b)); }
 inline unsigned ALIGN_UP(const LE32 &a, unsigned b) { return ALIGN_UP(unsigned(a), b); }
 
+// TODO later: introduce upx::umax() and upx::umin()
 inline unsigned UPX_MAX(unsigned a, const BE16 &b) { return UPX_MAX(a, unsigned(b)); }
 inline unsigned UPX_MAX(const BE16 &a, unsigned b) { return UPX_MAX(unsigned(a), b); }
 inline unsigned UPX_MIN(unsigned a, const BE16 &b) { return UPX_MIN(a, unsigned(b)); }
@@ -644,6 +785,47 @@ inline unsigned UPX_MIN(const LE32 &a, unsigned b) { return UPX_MIN(unsigned(a),
 // misc support
 **************************************************************************/
 
+// <type_traits> upx_is_integral; see conf.h
+#define TT_UPX_IS_INTEGRAL(T)                                                                      \
+    template <>                                                                                    \
+    struct upx_is_integral<T> : public std::true_type {};                                          \
+    template <>                                                                                    \
+    struct upx_is_integral<const T> : public std::true_type {};                                    \
+    template <>                                                                                    \
+    struct upx_is_integral<volatile T> : public std::true_type {};                                 \
+    template <>                                                                                    \
+    struct upx_is_integral<const volatile T> : public std::true_type {}
+TT_UPX_IS_INTEGRAL(BE16);
+TT_UPX_IS_INTEGRAL(BE32);
+TT_UPX_IS_INTEGRAL(BE64);
+TT_UPX_IS_INTEGRAL(LE16);
+TT_UPX_IS_INTEGRAL(LE32);
+TT_UPX_IS_INTEGRAL(LE64);
+#undef TT_UPX_IS_INTEGRAL
+
+// native types
+#if (ACC_ABI_BIG_ENDIAN)
+typedef BE16 NE16;
+typedef BE32 NE32;
+typedef BE64 NE64;
+#define ne16_compare        be16_compare
+#define ne32_compare        be32_compare
+#define ne64_compare        be64_compare
+#define ne16_compare_signed be16_compare_signed
+#define ne32_compare_signed be32_compare_signed
+#define ne64_compare_signed be64_compare_signed
+#else
+typedef LE16 NE16;
+typedef LE32 NE32;
+typedef LE64 NE64;
+#define ne16_compare        le16_compare
+#define ne32_compare        le32_compare
+#define ne64_compare        le64_compare
+#define ne16_compare_signed le16_compare_signed
+#define ne32_compare_signed le32_compare_signed
+#define ne64_compare_signed le64_compare_signed
+#endif
+
 // for use with qsort()
 extern "C" {
 int __acc_cdecl_qsort be16_compare(const void *, const void *);
@@ -700,14 +882,12 @@ namespace N_BELE_RTP {
 
 // util
 namespace N_BELE_CTP {
-inline const N_BELE_RTP::AbstractPolicy *getRTP(const BEPolicy * /*dummy*/) {
+inline const N_BELE_RTP::AbstractPolicy *getRTP(const BEPolicy * /*dummy*/) noexcept {
     return &N_BELE_RTP::be_policy;
 }
-inline const N_BELE_RTP::AbstractPolicy *getRTP(const LEPolicy * /*dummy*/) {
+inline const N_BELE_RTP::AbstractPolicy *getRTP(const LEPolicy * /*dummy*/) noexcept {
     return &N_BELE_RTP::le_policy;
 }
 } // namespace N_BELE_CTP
 
-#endif /* already included */
-
 /* vim:set ts=4 sw=4 et: */
